from dsdl.exception import DefineSyntaxError, DefineTypeError
from .utils import *
from dataclasses import dataclass
from typing import Optional


@dataclass()
class EleStruct:
    name: str
    type: str


class ParserField:
    def __init__(self, struct_name: str):
        self.TYPES_WITHOUT_PARS = [
            "Bool",
            "Num",
            "Int",
            "Str",
            "Coord",
            "Coord3D",
            "Interval",
            "BBox",
            "Polygon",
            "Image",
            "Video",
        ]
        self.TYPES_TIME = ["Date", "Time"]
        self.TYPES_LABEL = ["Label"]
        self.TYPES_LIST = ["List"]
        self.TYPES_ALL = (
            self.TYPES_WITHOUT_PARS
            + self.TYPES_TIME
            + self.TYPES_LABEL
            + self.TYPES_LIST
        )
        self.is_attr = set()
        self.optional = set()
        self.struct = struct_name  # field 中包含的struct，后续校验用

    def parse_list_filed(self, field_type: str, param_list: List[str]) -> str:
        """
        解析处理List类型的field
        """

        def sanitize_etype(val: str) -> str:
            """
            验证List类型中的etype是否存在（必须存在）且是否为合法类型
            """
            return self.pre_parse_struct_field(field_name=None, raw_field_type=val)

        def all_subclasses(cls):
            """
            返回某个类的所有子类：like[<class '__main__.Bar'>, <class '__main__.Baz'>...]
            """
            return cls.__subclasses__() + [
                g for s in cls.__subclasses__() for g in all_subclasses(s)
            ]

        def sanitize_ordered(val: str) -> str:
            if val.lower() in ["true", "false"]:
                if val.lower() == "true":
                    return "True"
                else:
                    return "False"
            else:
                raise DefineSyntaxError(
                    f"invalid value {val} in ordered of {field_type}: {','.join(param_list)}."
                )

        ele_type, ordered = None, None
        if len(param_list) == 2:
            ele_type = param_list[0].strip()
            ordered = param_list[1].strip()
        elif len(param_list) == 1:
            ele_type = param_list[0].strip()
        else:
            raise DefineSyntaxError(f"invalid parameters {param_list} in List.")

        if ele_type.startswith("etype"):
            # if ele_type like "etype=LocalObjectEntry[cdom=$objectdom]", split by "="
            # else if ele_type like "LocalObjectEntry[cdom=$objectdom]", do nothing.
            ele_type = ele_type.split("=", 1)
            if len(ele_type) == 2:
                if ele_type[0].strip() != "etype":
                    raise DefineSyntaxError(
                        f"List types must contains parameters `etype`."
                    )
                ele_type = ele_type[1].strip()
            else:
                raise DefineSyntaxError(
                    f"invalid parameters {', '.join(param_list)} in {field_type}."
                )

        res = field_type + "Field("
        if ele_type:
            ele_type = sanitize_etype(ele_type)
            res += "ele_type=" + ele_type
        else:
            raise DefineSyntaxError(f"List types must contains parameters `etype`.")
        if ordered:
            ordered = ordered.split("=")[-1]
            ordered = sanitize_ordered(ordered)
            res += ", ordered=" + ordered
        return res + ")"

    @staticmethod
    def parse_time_field(field_type: str, param_list: List[str]) -> str:
        """
        解析处理Time, Date类型的field
        """

        def sanitize_fmt(val: str) -> str:
            """
            Date, Time中对ftm部分的校验，是用来严格限制特定格式或者字符。
            防止yaml里有一些异常代码注入到生成的Python 代码里被执行起来。
            """
            val = val.strip("\"'")
            return f'"{val}"'

        param_dict = dict()
        for param in param_list:
            parts = param.split("=")
            parts = [i.strip() for i in parts]
            # 需要考虑参数省略的情况，因为dom经常省略
            if len(parts) == 2:
                field_para = parts[0]
                field_var = parts[1]
            elif len(parts) == 1:
                field_para = "fmt"
                field_var = parts[0]
            else:
                raise DefineSyntaxError(f"invalid parameters {param} in {field_type}.")

            if field_para != "fmt":
                raise DefineSyntaxError(
                    f"invalid parameters {field_para} in {field_type}."
                )

            if field_para in param_dict:
                raise ValueError(f"duplicated param {param} in {field_type}.")
            else:
                param_dict[field_para] = sanitize_fmt(field_var)

        return (
            field_type
            + "Field("
            + ", ".join([f"{k}={v}" for k, v in param_dict.items()])
            + ")"
        )

    @staticmethod
    def parse_label_field(field_type: str, param_list: List[str]) -> str:
        """
        解析处理Label类型的field
        """

        def sanitize_dom(val: str) -> str:
            """
            Label中对dom部分的校验，是用来严格限制特定格式或者字符。
            防止yaml里有一些异常代码注入到生成的Python 代码里被执行起来。
            """
            if not val.isidentifier():
                raise DefineSyntaxError(f"invalid dom: {val}")
            return val

        param_dict = dict()
        if not param_list:
            raise DefineSyntaxError(f"{field_type} must contains parameter `dom`")
        for param in param_list:
            parts = param.split("=")
            parts = [i.strip() for i in parts]
            # 需要考虑参数省略的情况，因为dom经常省略
            if len(parts) == 2:
                field_para = parts[0]
                field_var = parts[1]
            elif len(parts) == 1:
                field_para = "dom"
                field_var = parts[0]
            else:
                raise DefineSyntaxError(
                    f"invalid parameters {param_list} in {field_type}."
                )
            if field_para != "dom":
                raise DefineSyntaxError(
                    f"invalid parameters {field_para} in {field_type}."
                )

            if field_para in param_dict:
                raise ValueError(f"duplicated param {param} in {field_type}.")
            else:
                param_dict[field_para] = sanitize_dom(field_var)

        return (
            field_type
            + "Field("
            + ", ".join([f"{k}={v}" for k, v in param_dict.items()])
            + ")"
        )

    def parse_field(
        self,
        field_type: str,
        params_list: List[str] = None,
        attr_flag: bool = False,
        optional_flag: bool = False,
    ) -> str:
        """
        校验struct类型的每个字段的入口函数，对不同情况（Int,Image,List...）的字段进行校验并读入内存。
            raw_field_type: like: Int, Image, Label[dom=MyClassDom], List[List[Int], ordered = True], ....
        """
        if field_type in self.TYPES_WITHOUT_PARS:
            # 不带参数的不需要校验，直接可以转化为python代码（注意区分yaml中的模型部分到ORM的校验和通过ORM读入数据的校验，
            # 后者在dataset.py中定义，你脑子里想的Int也需要校验是数据的校验。）
            if params_list:
                raise DefineSyntaxError(f"{field_type} should not contains parameters.")
            field_type = field_type + "Field()"
        elif field_type in self.TYPES_LIST:
            # 带参数的List类型的字段的校验
            # （因为List类型比较特殊，里面可以包含List，Label等各种其他字段，涉及递归，所以单独拿出来）
            if not params_list:
                raise DefineSyntaxError(f"{field_type} must contains parameters.")
            field_type = self.parse_list_filed(field_type, params_list)
        elif field_type in self.TYPES_TIME:
            # 带参数的Date, Time类型的字段的校验
            field_type = self.parse_time_field(field_type, params_list)
        elif field_type in self.TYPES_LABEL:
            # 带参数的Label类型的字段的校验
            field_type = self.parse_label_field(field_type, params_list)
        else:
            # Struct类型的字段的校验
            if field_type not in self.struct:
                raise DefineTypeError(f"No type {field_type} in DSDL.")
            field_type = field_type + "()"
        if attr_flag:
            field_type = add_key_value_2_struct_field(field_type, "is_attr", True)
        if optional_flag:
            field_type = add_key_value_2_struct_field(field_type, "optional", True)
        return field_type

    def pre_parse_struct_field(
        self, field_name: Optional[str], raw_field_type: str
    ) -> str:
        # 将所有field可能都有的一些参数提取出来，类似is_attr、optional等
        raw_field_type = raw_field_type.strip()
        fixed_params = re.findall(r"\[(.*)\]", raw_field_type)
        if len(fixed_params) >= 2:
            raise DefineSyntaxError(f"Error in definition of {raw_field_type} in DSDL.")
        elif len(fixed_params) == 0:
            field_type = self.parse_field(field_type=raw_field_type)
        else:
            k_v_list = fixed_params[0]
            field_type = raw_field_type.replace("[" + k_v_list + "]", "")
            # below can split 'etype=LocalObjectEntry[cdom=COCO2017ClassDom, optional=True], optional=True' to
            # ['etype=LocalObjectEntry[cdom=COCO2017ClassDom', 'optional=True]', 'optional=True']
            k_v_list = re.split(r",\s*(?![^\[]*\])", k_v_list)
            k_v_list = [i.strip() for i in k_v_list]
            other_filed = set()
            attr_flag, optional_flag = False, False
            for k_v in k_v_list:
                if k_v.replace(" ", "") == "is_attr=True":
                    if field_name:
                        self.is_attr.add(field_name)
                    else:
                        attr_flag = True
                elif k_v.replace(" ", "") == "optional=True":
                    if field_name:
                        self.optional.add(field_name)
                    else:
                        optional_flag = True
                else:
                    other_filed.add(k_v)
            field_type = self.parse_field(
                field_type=field_type,
                params_list=list(other_filed),
                attr_flag=attr_flag,
                optional_flag=optional_flag,
            )
        return field_type



#!/usr/bin/env python3
#
# Copyright DSDL Team of OpenDatalab. All rights reserved.
#

from .parser import parse
from .dataset import Dataset, ImageVisualizePipeline
from .__version__ import __version__



import click
from typing import Sequence, Union
import os
try:
    from yaml import CSafeLoader as YAMLSafeLoader
except ImportError:
    from yaml import SafeLoader as YAMLSafeLoader
from yaml import load as yaml_load
import json



class OptionEatAll(click.Option):
    """
    implemented by https://stackoverflow.com/users/7311767/stephen-rauch in https://stackoverflow.com/questions/48391777/nargs-equivalent-for-options-in-click
    """

    def __init__(self, *args, **kwargs):
        self.save_other_options = kwargs.pop('save_other_options', True)
        nargs = kwargs.pop('nargs', -1)
        assert nargs == -1, 'nargs, if set, must be -1 not {}'.format(nargs)
        super(OptionEatAll, self).__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._eat_all_parser = None

    def add_to_parser(self, parser, ctx):

        def parser_process(value, state):
            # method to hook to the parser.process
            done = False
            value = [value]
            if self.save_other_options:
                # grab everything up to the next option
                while state.rargs and not done:
                    for prefix in self._eat_all_parser.prefixes:
                        if state.rargs[0].startswith(prefix):
                            done = True
                    if not done:
                        value.append(state.rargs.pop(0))
            else:
                # grab everything remaining
                value += state.rargs
                state.rargs[:] = []
            value = tuple(value)

            # call the actual process
            self._previous_parser_process(value, state)

        retval = super(OptionEatAll, self).add_to_parser(parser, ctx)
        for name in self.opts:
            our_parser = parser._long_opt.get(name) or parser._short_opt.get(name)
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break
        return retval


YAML_VALID_SUFFIX = ('.yaml', '.YAML')
JSON_VALID_SUFFIX = ('.json', '.JSON')
VALID_SUFFIX = YAML_VALID_SUFFIX + JSON_VALID_SUFFIX


def load_samples(dsdl_path: str, path: Union[str, Sequence[str]]):
    samples = []
    paths = []
    dsdl_dir = os.path.split(dsdl_path)[0]
    if isinstance(path, str):
        path = os.path.join(dsdl_dir, path)
        if os.path.isdir(path):
            paths = [os.path.join(path, _) for _ in os.listdir(path) if _.endswith(VALID_SUFFIX)]
        elif os.path.isfile(path):
            if path.endswith(VALID_SUFFIX):
                paths = [path]
    elif isinstance(path, (list, tuple)):
        paths = [os.path.join(dsdl_dir, _) for _ in path if os.path.isfile(_) and _.endswith(VALID_SUFFIX)]
    for p in paths:
        if p.endswith(YAML_VALID_SUFFIX):
            with open(p, "r") as f:
                data = yaml_load(f, YAMLSafeLoader)['samples']
            samples.extend(data)
        else:
            with open(p, "r") as f:
                data = json.load(f)['samples']
            samples.extend(data)
    return samples





import numpy as np
from PIL import Image, ExifTags
from typing import Tuple, Any
import io
from ..exception import FileReadError


def get_image_rotation(image: Image) -> int:
    """Get the rotation degree from the image file's exif message.

    Arguments:
        image: A PIL.Image object.

     Returns:
         The degree read from image's exif message.
    """
    try:
        for k, v in ExifTags.TAGS.items():
            if v == "Orientation":
                break
        orientation_degree_map = {3: 180, 6: 270, 8: 90}
        return orientation_degree_map.get(image._getexif()[k], 0)
    except Exception:
        return 0


def bytes_to_numpy(bytes_: io.BytesIO) -> np.ndarray:  # type: ignore[type-arg]
    """
    Transfer bytes into numpy array.

    Arguments:
        bytes_: The bytes to transfer.

    Raises:
        FileReadError: When `bytes_` cannot be loaded as an image.

    Returns:
        The transferred numpy array.
    """
    try:
        image = Image.open(bytes_)
    except Exception as e:
        raise FileReadError(f"Failed to convert bytes to an array. {e}") from None
    rotation = get_image_rotation(image)
    if rotation:
        image = image.rotate(rotation, expand=True)
    if image.mode == "RGB":
        shape: Tuple[int, ...] = (image.size[1], image.size[0], 3)
        dtype: Any = np.uint8
    elif image.mode == "RGBA":
        shape: Tuple[int, ...] = (image.size[1], image.size[0], 4)
        dtype: Any = np.uint8
    elif image.mode == "P":
        shape = (image.size[1], image.size[0])
        dtype = np.uint8
    elif image.mode == "I":
        shape = (image.size[1], image.size[0])
        dtype = np.int32
    elif image.mode == "L":
        shape = (image.size[1], image.size[0])
        dtype = np.uint8
    elif image.mode == "LA":
        shape = (image.size[1], image.size[0], 2)
        dtype = np.uint8
    else:
        raise FileReadError("Currently unsupported image type")
    image_ = np.array(image.getdata(), dtype=dtype).reshape(*shape)
    return image_



class DuplicateDefineWarning(SyntaxWarning):
    pass


class InvalidLabelWarning(SyntaxWarning):
    pass


class FieldNotFoundWarning(SyntaxWarning):
    pass


class ClassHasDefinedWarning(SyntaxWarning):
    pass



# Generated by the dsdl parser. DO NOT EDIT!

from dsdl.types import *
from enum import Enum, unique


class COCO2017ClassFatherDom(ClassDomain):
    Classes = [
        Label("food"),
        Label("tool"),
    ]


class COCO2017ClassFatherDom1(ClassDomain):
    Classes = [
        Label("fruit", supercategories=[COCO2017ClassFatherDom.get_label("food")]),
        Label("sports tool", supercategories=[COCO2017ClassFatherDom.get_label("tool")]),
    ]


class COCO2017ClassDom(ClassDomain):
    Classes = [
        Label("airplane", supercategories=[COCO2017ClassFatherDom1.get_label("sports tool")]),
        Label("apple", supercategories=[COCO2017ClassFatherDom1.get_label("fruit")]),
        Label("backpack", supercategories=[COCO2017ClassFatherDom1.get_label("")]),
        Label("banana", supercategories=[COCO2017ClassFatherDom1.get_label("fruit")]),
        Label("baseball bat", supercategories=[COCO2017ClassFatherDom1.get_label("sports tool")]),
        Label("baseball glove", supercategories=[COCO2017ClassFatherDom1.get_label("sports tool")]),
    ]


class LocalObjectEntry(Struct):
    bbox = BBoxField()
    label = LabelField(dom=COCO2017ClassDom, optional=True)


class SceneAndObjectSample(Struct):
    image = ImageField()
    sclabel = LabelField(dom=COCO2017ClassFatherDom, optional=True)
    objects = ListField(ele_type=LocalObjectEntry())



from typing import TypeVar, List
import numpy as np
from PIL import ImageDraw
from .base_geometry import BaseGeometry

_ELE_TYPE = TypeVar("_ELE_TYPE", int, float)


class BBox(BaseGeometry):

    def __init__(
            self,
            x: _ELE_TYPE,
            y: _ELE_TYPE,
            width: _ELE_TYPE,
            height: _ELE_TYPE
    ):
        self._data = [x, y, width, height]

    @property
    def x(self) -> _ELE_TYPE:
        return self._data[0]

    @property
    def y(self) -> _ELE_TYPE:
        return self._data[1]

    @property
    def width(self) -> _ELE_TYPE:
        return self._data[2]

    @property
    def height(self) -> _ELE_TYPE:
        return self._data[3]

    @property
    def xmin(self) -> _ELE_TYPE:
        return self._data[0]

    @property
    def ymin(self) -> _ELE_TYPE:
        return self._data[1]

    @property
    def xmax(self) -> _ELE_TYPE:
        return self._data[0] + self._data[2]

    @property
    def ymax(self) -> _ELE_TYPE:
        return self._data[1] + self._data[3]

    @property
    def area(self) -> _ELE_TYPE:
        return self.width * self.height

    @property
    def xyxy(self) -> List[_ELE_TYPE]:
        return [self.xmin, self.ymin, self.xmax, self.ymax]

    @property
    def xywh(self) -> List[_ELE_TYPE]:
        return [self.xmin, self.ymin, self.width, self.height]
    @property
    def openmmlabformat(self) -> List[_ELE_TYPE]:
        return [self.xmin, self.ymin, self.xmax, self.ymax]

    def to_int(self):
        self._data = [int(_) for _ in self._data]

    def to_float(self):
        self._data = [float(_) for _ in self._data]

    def visualize(self, image, palette, **kwargs):
        draw_obj = ImageDraw.Draw(image)
        color = (0, 255, 0)
        if "label" in kwargs:
            for label in kwargs["label"].values():
                if label.category_name not in palette:
                    palette[label.category_name] = tuple(np.random.randint(0, 255, size=[3]))

                color = palette[label.category_name]
        draw_obj.rectangle(self.xyxy, outline=(*color, 255), width=2)
        del draw_obj
        return image

    def __repr__(self):
        return str(self.xyxy)



from .field import Field
from .struct import Struct
from ..exception import ValidationError


class BoolField(Field):
    def validate(self, value):
        if value not in [True, False]:
            raise ValidationError(f"expect True/False, got {value}")
        return value


class IntField(Field):
    def validate(self, value):
        try:
            return int(value)
        except (ValueError, TypeError) as _:
            raise ValidationError(f"expect int, got {value}")


class NumField(Field):
    def validate(self, value):
        try:
            return float(value)
        except (ValueError, TypeError) as _:
            raise ValidationError(f"expect num, got {value}")


class StrField(Field):
    def validate(self, value):
        try:
            return "" + value
        except TypeError as _:
            raise ValidationError(f"expect str, got {value}")


class ListField(Field):
    def __init__(self, ele_type, ordered=False, *args, **kwargs):
        self.ordered = ordered
        self.ele_type = ele_type
        self.file_reader = None
        super().__init__(*args, **kwargs)

    def validate(self, value):
        if hasattr(self.ele_type, "set_file_reader"):
            self.ele_type.set_file_reader(self.file_reader)
        if isinstance(self.ele_type, Field):
            value = [self.ele_type.validate(item) for item in value]
        elif isinstance(self.ele_type, Struct):
            value = [self.ele_type.__class__(file_reader=self.file_reader, **item) for item in value]
        return value

    def set_file_reader(self, file_reader):
        self.file_reader = file_reader



from .label import Label
from .class_domain import ClassDomain
from .base_geometry import BaseGeometry
from typing import List
from ..exception import ClassNotFoundError
from PIL import ImageDraw
import numpy as np


class Coord2D(BaseGeometry):

    def __init__(self, x: float, y: float, visiable: int, label: Label):
        self._x = x
        self._y = y
        self._visiable = visiable
        self._label = label

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def point(self):
        return [self._x, self._y]

    @property
    def value(self):
        return [self._x, self._y, self._visiable]

    @property
    def visiable(self):
        return self._visiable

    @property
    def label(self):
        return self._label

    @property
    def class_domain(self):
        return self._label.class_domain

    @property
    def name(self):
        return self._label.name


class KeyPoints(BaseGeometry):
    def __init__(self, keypoints: List[Coord2D], domain: ClassDomain):
        self._keypoints = keypoints
        self._dom = domain

    @property
    def value(self):
        return [_.value for _ in self._keypoints]

    @property
    def points(self):
        return [_.point for _ in self._keypoints]

    @property
    def visables(self):
        return [_.visiable for _ in self._keypoints]

    @property
    def keypoints(self):
        return self._keypoints

    @property
    def class_domain(self):
        return self._dom

    @property
    def names(self):
        return [_.name for _ in self._keypoints]

    def __getitem__(self, item):
        assert isinstance(item, (str, int)), "The index must be str or int type value."
        if isinstance(item, int):
            return self._keypoints[item]
        elif isinstance(item, str):
            for ind, label_name in enumerate(self._dom.get_label_names()):
                if label_name == item:
                    return self._keypoints[ind]
        raise ClassNotFoundError(f"Category '{item}' not defined in domain {self._dom.__name__}.")

    def visualize(self, image, palette, **kwargs):
        draw_obj = ImageDraw.Draw(image)
        line_color = (0, 255, 0)  # green
        point_radius = 3
        skeleton = self._dom.get_attribute("Skeleton")
        point_pairs = skeleton.get_point_pairs(self)
        for point_pair in point_pairs:
            p1, p2 = point_pair[:2]
            if p1.visiable > 0 and p2.visiable:
                draw_obj.line([*p1.point, *p2.point], width=2, fill=(*line_color, 255))
        for point in self._keypoints:
            if point.visiable > 0:
                label_ = point.label.category_name
                if label_ not in palette:
                    palette[label_] = tuple(np.random.randint(0, 255, size=[3]))
                point_color = palette[label_]
                x, y = point.point
                draw_obj.ellipse((x - point_radius, y - point_radius, x + point_radius, y + point_radius),
                                 fill=(*point_color, 255))
        del draw_obj
        return image



from typing import List
import numpy as np
from PIL import ImageDraw, Image
from .base_geometry import BaseGeometry


class PolygonItem(BaseGeometry):

    def __init__(
            self,
            points: List[List[float]]
    ):
        self._data = points

    @property
    def points(self) -> List[List[float]]:
        return self._data

    @property
    def points_x(self) -> List[float]:
        return [_[0] for _ in self._data]

    @property
    def points_y(self) -> List[float]:
        return [_[1] for _ in self._data]

    @property
    def point_for_draw(self) -> List[int]:
        p = min(self._data, key=lambda x: x[0] + x[1])
        p = [int(_) for _ in p]
        return p

    @property
    def openmmlabformat(self) -> List[float]:
        return self._flatten()

    def to_tuple(self):
        return tuple([(_[0], _[1]) for _ in self._data])

    def _flatten(self) -> List[float]:
        return [_ for point in self._data for _ in point]

    def __repr__(self):
        return str(self._data)


class Polygon(BaseGeometry):

    def __init__(self, polygons: List[PolygonItem]):
        self._data = polygons

    @property
    def polygons(self):
        return self._data

    @property
    def openmmlabformat(self) -> List[List[float]]:
        return [_.openmmlabformat for _ in self._data]

    @property
    def point_for_draw(self) -> [int, int]:
        p = min(self._data, key=lambda x: x.point_for_draw[0] + x.point_for_draw[1])
        p = [int(_) for _ in p.point_for_draw]
        return p

    def visualize(self, image, palette, **kwargs):
        color = (0, 255, 0)
        if "label" in kwargs:
            for label in kwargs["label"].values():
                if label.category_name not in palette:
                    palette[label.category_name] = tuple(np.random.randint(0, 255, size=[3]))

                color = palette[label.category_name]

        poly = Image.new('RGBA', image.size[:2])
        pdraw = ImageDraw.Draw(poly)
        for polygon_item in self.polygons:
            pdraw.polygon(polygon_item.to_tuple(), fill=(*color, 127), outline=(*color, 255))
        image.paste(poly, mask=poly)
        del pdraw
        return image

    def __repr__(self):
        return str(self._data)



from .box import BBox
from .label import Label, LabelList
from .media import ImageMedia
from .polygon import Polygon, PolygonItem
from .attrbutes import Attributes
from .segmap import SegmentationMap
from .keypoint import Coord2D, KeyPoints
from .registry import STRUCT, CLASSDOMAIN, LABEL
from .class_domain import ClassDomain

__all__ = [
    "BBox",
    "Label",
    "ImageMedia",
    "LabelList",
    "Polygon",
    "PolygonItem",
    "Attributes",
    "SegmentationMap",
    "Coord2D",
    "STRUCT",
    "CLASSDOMAIN",
    "LABEL",
    "ClassDomain",
    "KeyPoints",
]



from .base_geometry import BaseGeometry
from .utils import bytes_to_numpy
from ..exception import FileReadError
import numpy as np
from PIL import Image
import io
from .label import LabelList
from .box import BBox


class SegmentationMap(BaseGeometry):
    """
    A Geometry class for semantic segmentation map.
    """

    def __init__(self, location, file_reader, dom):
        self._loc = location
        self._reader = file_reader
        self._dom = dom

    @property
    def class_domain(self):
        return self._dom

    @property
    def location(self):
        return self._loc

    def to_bytes(self):
        """
        turn SegmentationMap object to bytes
        """
        return io.BytesIO(self._reader.read())

    def to_image(self):
        """
        turn SegmentationMap object to PIL.Image
        """
        try:
            img = Image.open(self.to_bytes())
        except Exception as e:
            raise FileReadError(f"Failed to convert bytes to an array. {e}") from None
        return img

    def to_array(self):
        """
        turn SegmentationMap object to numpy.ndarray
        """
        return bytes_to_numpy(self.to_bytes())

    def visualize(self, image, palette, **kwargs):
        seg = self.to_array()
        color_seg = np.zeros((seg.shape[0], seg.shape[1], 3), dtype=np.uint8)
        category_ids = np.unique(seg)
        label_lst = []
        for category_id in category_ids:
            label = self._dom.get_label(int(category_id))
            category_name = label.category_name
            if category_name not in palette:
                palette[category_name] = tuple(np.random.randint(0, 255, size=[3]))
            label_lst.append(label)
            color_seg[seg == category_id, :] = np.array(palette[category_name])
        overlay = Image.fromarray(color_seg).convert("RGBA")
        overlayed = Image.blend(image, overlay, 0.5)
        LabelList(label_lst).visualize(image=overlayed, palette=palette, bbox={"temp": BBox(0, 0, 0, 0)})
        return overlayed

    def __repr__(self):
        return f"path:{self.location}, class domain: {self._dom.__name__}"



from .struct import Struct
from .generic import StrField, IntField, BoolField, NumField, ListField
from .special import (
    LabelField,
    CoordField,
    Coord3DField,
    IntervalField,
    BBoxField,
    PolygonField,
    DateField,
    TimeField,
    KeypointField,
)
from .unstructure import ImageField, SegMapField

__all__ = [
    "Struct",
    "StrField",
    "IntField",
    "BoolField",
    "NumField",
    "ImageField",
    "LabelField",
    "ListField",
    "CoordField",
    "Coord3DField",
    "IntervalField",
    "BBoxField",
    "PolygonField",
    "DateField",
    "TimeField",
    "SegMapField",
    "KeypointField",
]



class Parser:
    """
    {
        $image: {img: reader_obj},
        $list: {objects: [{bbox: {bbox: xxx}, label: {label: xxx}, extra: {}}]}
        $extra: {}
    }
    -->
    {img: xxx
    objects: []
    }
    """

    @staticmethod
    def flatten_sample(sample, field_name, parse_method=lambda _: _):
        result_dic = {}
        prefix = "."
        Parser._helper(sample, field_name, result_dic, prefix, parse_method)
        return result_dic

    @staticmethod
    def _helper(sample, field_name, result_dic, prefix=".", parse_method=lambda _: _):

        if isinstance(sample, dict):
            for field_type in sample:
                if field_type == field_name:
                    for key, value in sample[field_type].items():
                        k_ = f"{prefix}/{key}"
                        result_dic[k_] = parse_method(value)
                if field_type == "$list":
                    for key, value in sample[field_type].items():
                        k_ = f"{prefix}/{key}"
                        Parser._helper(value, field_name, result_dic, prefix=k_, parse_method=parse_method)

        elif isinstance(sample, list):
            for id_, item in enumerate(sample):
                k_ = f"{prefix}/{id_}"
                Parser._helper(item, field_name, result_dic, prefix=k_, parse_method=parse_method)


if __name__ == '__main__':
    sample = {
        "$image": {"img2": None, "img1": None},
        "$list": {"objects": [{"$image": {"img1": None}, "$bbox": {"box": None}}],
                  "object2": [{"$bbox": {"box": None}}, {"$bbox": {"box": 1}}], }
    }

    from collections import defaultdict

    dic = defaultdict(defaultdict)



from dsdl.dataset import Dataset, ImageVisualizePipeline, Util
import click
import numpy as np
from random import randint
import cv2
import os

try:
    from yaml import CSafeLoader as YAMLSafeLoader
except ImportError:
    from yaml import SafeLoader as YAMLSafeLoader
from yaml import load as yaml_load
from .commons import OptionEatAll, load_samples
from ..parser import dsdl_parse


@click.command()
@click.option("-y", "--yaml", "dsdl_yaml", type=str, required=True, help="the path of dsdl yaml file")
@click.option("-c", "--config", "config", type=str, required=True, help="the path of the config file")
@click.option("-l", "--location", "location", type=click.Choice(["local", "ali-oss"]), required=True,
              help="the path of the config file")
@click.option("-n", "--num", "num", type=int, default=5, help="how many samples sampled from the dataset")
@click.option("-r", "--random", is_flag=True, help="whether to sample randomly")
@click.option("-v", "--visualize", is_flag=True, help="whether to visualize the sample selected")
@click.option("-f", "--fields", cls=OptionEatAll, type=str, help="the task to visualize")
@click.option("-t", "--task", type=str, help="the task to visualize")
@click.option("-p", "--position", type=str, required=False, help='the directory of dsdl define file')
@click.option("-m", "--multistage", is_flag=True, help="whether to use the generated python file")
def view(dsdl_yaml, config, location, num, random, visualize, fields, task, position, multistage):
    with open(dsdl_yaml, "r") as f:
        dsdl_info = yaml_load(f, Loader=YAMLSafeLoader)['data']
        sample_type = dsdl_info['sample-type']
        sample_path = dsdl_info["sample-path"]
        if sample_path == "$local" or sample_path == "local":
            samples = dsdl_info['samples']
        else:
            samples = load_samples(dsdl_yaml, sample_path)
    if multistage:
        dsdl_py = os.path.splitext(dsdl_yaml)[0] + ".py"
        with open(dsdl_py, encoding='utf-8') as dsdl_file:
            exec(dsdl_file.read(), {})
    else:
        if position:
            dsdl_py = dsdl_parse(dsdl_yaml, position)
        else:
            dsdl_py = dsdl_parse(dsdl_yaml)
        exec(dsdl_py, {})
    config_dic = {}
    with open(config, encoding='utf-8') as config_file:
        exec(config_file.read(), config_dic)
    location_config = config_dic["local" if location == "local" else "ali_oss"]
    dataset = Dataset(samples, sample_type, location_config)

    palette = {}
    if task:
        assert task in ["detection",
                        "segmentation",
                        "classification"], "invalid task, you can only choose in ['detection', 'segmentation', 'classification']"
        if task == "classification":
            fields = ["image", "label"]
        elif task == "detection":
            fields = ["image", "label", "bbox", "polygon", "attributes", "keypoint"]
        elif task == "segmentation":
            fields = ["image", "segmap", "attributes"]
        else:
            fields = ["image", "label", "attributes"]
    else:
        if fields is None:
            fields = []
        else:
            local_dic = {}
            exec(f"f = list({fields})", {}, local_dic)
            fields = local_dic['f']
        fields = list(set(fields + ["image", "attributes"]))
    fields = [_.lower() for _ in fields]

    num = min(num, len(dataset))

    if not random:
        indices = list(range(num))
    else:
        indices = [randint(0, len(dataset) - 1) for _ in range(num)]

    samples = []
    for ind in indices:
        samples.append(ImageVisualizePipeline(sample=dataset[ind], palette=palette, field_list=fields))

    print(Util.format_sample([s.format() for s in samples]))
    # 将读取的样本进行可视化
    if visualize:
        for sample in samples:
            vis_sample = sample.visualize()
            for vis_name, vis_item in vis_sample.items():
                cv2.imshow(vis_name, cv2.cvtColor(np.array(vis_item), cv2.COLOR_BGR2RGB))
                cv2.waitKey(0)



class Field:
    def __init__(self, is_attr=False, optional=False):
        self._is_attr = is_attr
        self._optional = optional

    @property
    def is_attr(self):
        return self._is_attr

    @property
    def is_optional(self):
        return self._optional

    def __str__(self):
        return "<%s>" % self.__class__.__name__

    def validate(self, value):
        """
        Validate value and raise ValidationError if necessary.
        """
        return value



from contextlib import contextmanager


class BaseFileReader:

    def __init__(self, working_dir=""):
        self.working_dir = working_dir

    @contextmanager
    def load(self, file):
        raise NotImplementedError



from .base_dataset import Dataset
from .utils.commons import Util
from .demo_dataset import DemoDataset
from .utils.visualizer import ImageVisualizePipeline

__all__ = [
    "Dataset",
    "DemoDataset",
    "ImageVisualizePipeline",
    "Util",
]



import os
from contextlib import contextmanager
from .base import BaseFileReader


class LocalFileReader(BaseFileReader):

    @contextmanager
    def load(self, file):
        fp = os.path.join(self.working_dir, file)
        f = open(fp, "rb")
        try:
            yield f
        finally:
            f.close()



import click

from dsdl import parse
from dsdl.tools import view


@click.group()
def cli():
    pass


# parser解析器的入口：用来解析yaml里面的模型和标签部分。
cli.add_command(parse)
cli.add_command(view)

if __name__ == "__main__":
    cli()



from .field import Field
from ..geometry import BBox, Polygon, PolygonItem, Coord2D, KeyPoints
from ..exception import ValidationError
from datetime import date, time, datetime


def validate_list_of_number(value, size_limit, item_type):
    if type(value) is not list:
        raise ValidationError(f"expect list of num, got {value}")
    if len(value) != size_limit:
        raise ValidationError(f"expect size of list is {size_limit}, got {len(value)}")
    try:
        return [item_type(item) for item in value]
    except (TypeError, ValueError) as _:
        raise ValidationError(f"expect type of list item is float, got {value}")


class CoordField(Field):
    def validate(self, value):
        return validate_list_of_number(value, 2, float)


class Coord3DField(Field):
    def validate(self, value):
        return validate_list_of_number(value, 3, float)


class IntervalField(Field):
    def validate(self, value):
        value = validate_list_of_number(value, 2, float)
        if value[0] > value[1]:
            raise ValidationError(
                f"expect |begin| less than or equal to |end|, got {value}"
            )
        return value


class BBoxField(Field):
    def validate(self, value):
        return BBox(*validate_list_of_number(value, 4, float))


class PolygonField(Field):
    def validate(self, value):
        polygon_lst = []
        for idx, points in enumerate(value):
            for point in points:
                validate_list_of_number(point, 2, float)
            polygon_lst.append(PolygonItem(points))
        return Polygon(polygon_lst)


class KeypointField(Field):

    def __init__(self, dom, *args, **kwargs):
        super(KeypointField, self).__init__(*args, **kwargs)
        self.dom = dom

    def validate(self, value):
        value = validate_list_of_number(value, len(self.dom), list)
        keypoints = []
        for class_ind, p in enumerate(value, start=1):
            p = validate_list_of_number(p, 3, float)
            label = self.dom.get_label(class_ind)
            coord2d = Coord2D(x=p[0], y=p[1], visiable=int(p[2]), label=label)
            keypoints.append(coord2d)

        return KeyPoints(keypoints=keypoints, domain=self.dom)


class LabelField(Field):
    def __init__(self, dom, *args, **kwargs):
        super(LabelField, self).__init__(*args, **kwargs)
        self.dom = dom

    def validate(self, value):
        try:
            if isinstance(value, (int, str)):
                return self.dom.get_label(value)
            else:
                raise TypeError("invalid class label type.")
        except:
            raise RuntimeError(f"The label {value} is not valid.")


class DateField(Field):
    def __init__(self, fmt: str = "", *args, **kwargs):
        super(DateField, self).__init__(*args, **kwargs)
        self.fmt = fmt

    def validate(self, value):
        if self.fmt == "":
            return date.fromisoformat(value)
        return datetime.strptime(value, self.fmt).date()


class TimeField(Field):
    def __init__(self, fmt: str = "", *args, **kwargs):
        super(TimeField, self).__init__(*args, **kwargs)
        self.fmt = fmt

    def validate(self, value):
        if self.fmt == "":
            return time.fromisoformat(value)
        return datetime.strptime(value, self.fmt).time()



from .base_dataset import Dataset

try:
    from yaml import CSafeLoader as YAMLSafeLoader
except ImportError:
    from yaml import SafeLoader as YAMLSafeLoader


class DemoDataset(Dataset):

    def __getitem__(self, idx):
        return self.sample_list[idx]

    def _parse_struct(self, sample):
        return sample



import oss2
import os
from contextlib import contextmanager
from .base import BaseFileReader


class AliOSSFileReader(BaseFileReader):
    """
    该类的作用为读取 阿里云OSS上面的文件
    """

    def __init__(self, working_dir, bucket_name, access_key_id, access_key_secret, endpoint):
        super().__init__(working_dir)
        auth = oss2.Auth(access_key_id, access_key_secret)
        self.bucket = oss2.Bucket(auth, endpoint, bucket_name)

    @contextmanager
    def load(self, file):
        fp = os.path.join(self.working_dir, file)
        object_stream = self.bucket.get_object(fp)
        try:
            yield object_stream
        finally:
            if object_stream.client_crc != object_stream.server_crc:
                print("The CRC checksum between client and server is inconsistent!")



from ..exception import ClassNotFoundError
from ..warning import ClassHasDefinedWarning


class Registry:
    def __init__(self, name):
        self._name = name
        self._map = {}

    @property
    def name(self):
        return self._name

    def register(self, name, cls):
        # NOTICE: Since this method is called when models are imported,
        # it cannot perform imports because of the risk of import loops.
        if name in self._map:
            # raise ClassHasDefinedError
            ClassHasDefinedWarning(f"Class '{name}' has been registered, it will be replaced by this updated one.")
        self._map[name] = cls

    def get(self, name):
        if name not in self._map:
            raise ClassNotFoundError(f"Class '{name}' is not defined.")
        return self._map[name]


STRUCT = Registry("struct")
CLASSDOMAIN = Registry("class domain")


class LabelRegistry:
    def __init__(self):
        self._labels = {}

    def registry(self, label):
        registry_name = label.registry_name
        if registry_name in self._labels:
            # raise ClassHasDefinedError(f"The label {label.registry_name} has been registered.")
            ClassHasDefinedWarning(
                f"The label {label.registry_name} has been registered, it will be replaced bt this updated one.")
        self._labels[registry_name] = label

    def get(self, registry_name):
        if registry_name not in self._labels:
            raise ClassNotFoundError
        return self._labels[registry_name]

    def __contains__(self, item):
        if isinstance(item, str):
            return item in self._labels
        elif hasattr(item, "registry_name"):
            return item.registry_name in self._labels
        else:
            return False


LABEL = LabelRegistry()



from .utils import *
from dsdl.exception import DefineSyntaxError
from dataclasses import dataclass, field
import re
from typing import Optional, Tuple


@dataclass()
class SuperCategories:
    class_dom: str
    label_value: str


@dataclass()
class EleClassDetail:
    label_value: str
    super_categories: Optional[List[SuperCategories]] = field(default_factory=list)


@dataclass()
class EleClass:
    label_value: str
    super_categories: Optional[List[str]] = field(default_factory=list)


class ParserClass:
    def __init__(self, class_name: str, class_value: List[str]):
        self.class_name, self.super_class_list = self._parse_class_name(
            raw_name=class_name
        )
        self.class_field_detail = self._parse_class_value(class_value=class_value)
        self.class_field = self._parse_super_categories()

    @staticmethod
    def _parse_class_name(raw_name: str) -> Tuple[str, Optional[List[str]]]:
        super_class = re.findall(r"\[(.*)\]", raw_name)
        if len(super_class) >= 2:
            raise DefineSyntaxError(f"Error in definition of {raw_name} in DSDL.")
        elif len(super_class) == 0:
            class_name = raw_name
            super_class_list = []
        else:
            super_class_list = super_class[0]
            class_name = raw_name.replace("[" + super_class_list + "]", "")
            super_class_list = super_class_list.strip().split(",")
        return class_name, super_class_list

    def _parse_class_value(self, class_value: List[str]) -> List[EleClassDetail]:
        class_field = list()
        if self.super_class_list:
            for value in class_value:
                super_class_value = re.findall(r"\[(.*?)\]", value)
                assert len(super_class_value) == len(
                    self.super_class_list
                ), f"length of {value} must equal to nums of super_categories."

                label = value.split("[")[0]
                ele_class = EleClassDetail(label_value=label)
                for super_class_name, label_value in zip(
                    self.super_class_list, super_class_value
                ):
                    if label_value:
                        temp_super = SuperCategories(
                            class_dom=super_class_name, label_value=label_value
                        )
                        ele_class.super_categories.append(temp_super)
                class_field.append(ele_class)
        else:
            for value in class_value:
                super_class_value = re.findall(r"\[(.*?)\]", value)
                assert (
                    len(super_class_value) == 0
                ), f"length of {value} must equal to nums of super_categories."
                ele_class = EleClassDetail(
                    label_value=value.strip(), super_categories=None
                )
                class_field.append(ele_class)
        return class_field

    def _parse_super_categories(self) -> List[EleClass]:
        class_field = list()
        for ele in self.class_field_detail:
            ele_class = EleClass(label_value=ele.label_value)
            if ele.super_categories:
                ele_class.super_categories = list()
                for super_label in ele.super_categories:
                    temp = (
                        super_label.class_dom
                        + '.get_label("'
                        + super_label.label_value
                        + '")'
                    )
                    ele_class.super_categories.append(temp)
            else:
                ele_class.super_categories = None
            class_field.append(ele_class)
        return class_field



import io
from PIL import Image
from ..exception import FileReadError
from .base_geometry import BaseGeometry
from .utils import bytes_to_numpy


class ImageMedia(BaseGeometry):

    def __init__(self, location, file_reader):
        self._loc = location
        self._reader = file_reader

    @property
    def location(self):
        return self._loc

    def to_bytes(self):
        """
        turn ImageMedia object to bytes
        """
        return io.BytesIO(self._reader.read())

    def to_image(self):
        """
        turn ImageMedia object to PIL.Image
        """
        try:
            img = Image.open(self.to_bytes())
        except Exception as e:
            raise FileReadError(f"Failed to convert bytes to an array. {e}") from None
        return img

    def to_array(self):
        """
        turn ImageMedia object to numpy.ndarray
        """
        return bytes_to_numpy(self.to_bytes())

    def __repr__(self):
        return f"path:{self.location}"



import re
from dsdl.types.field import Field
from prettytable import PrettyTable


class Util:
    @staticmethod
    def extract_key(field_obj: Field):
        field_cls_name = field_obj.__class__.__name__
        return "$" + field_cls_name.replace("Field", "").lower()

    @staticmethod
    def get_first_item(dic):
        if len(dic) == 0:
            return None
        for k in dic:
            return {"key": k, "value": dic[k]}

    @staticmethod
    def extract_sample_type(sample_type):
        c_dom = re.findall(r"\[(.*?)\]", sample_type)
        if c_dom:
            return sample_type.replace("[" + c_dom[0] + "]", "", 1).strip()
        else:
            return sample_type

    @staticmethod
    def format_sample(samples):
        """
        该方法的作用就是将parse_struct方法的返回值（单个样本）写到一个列表中，方便在命令行中展示样本的基本信息
        """
        table = PrettyTable()
        table.title = "Samples"
        table.field_names = ["sample idx", "media", "annotation"]
        table._max_width = {"media": 20, "annotation": 50}

        for idx, sample in enumerate(samples, start=1):
            anns = [s["ann"] for s in sample]
            medias = [s["media"] for s in sample]
            processed_anns = [[Util.strip(dic) for dic in lst] for lst in anns]
            if len(medias) == 1:
                processed_anns = processed_anns[0]
                medias = medias[0]
            table.add_row([idx, medias, processed_anns])
        return table

    @staticmethod
    def strip(dic):
        result_dic = {}
        for sub_dic in dic.values():
            for k in sub_dic:
                result_dic[k] = sub_dic[k]
        return result_dic



import numpy as np
from PIL import ImageFont, ImageDraw
from .registry import CLASSDOMAIN
from .base_geometry import BaseGeometry


class Label(BaseGeometry):
    def __init__(self, name, supercategories=(), domain_name=None):
        self._name = name
        self._supercategories = [_ for _ in supercategories if isinstance(_, Label)]
        self._domain_name = domain_name

    @property
    def supercategories(self):
        return self._supercategories

    @property
    def supercategories_names(self):
        return [_.name for _ in self._supercategories]

    @property
    def parent_names(self):
        return [_.name for _ in self._supercategories]

    @property
    def domain_name(self):
        return self._domain_name

    @property
    def parents(self):
        return self._supercategories

    @property
    def name(self):
        return self._name

    @property
    def registry_name(self):
        return f"{self._domain_name}__{self._name}"

    def set_domain(self, domain_name):
        if self._domain_name is None:
            self._domain_name = domain_name

    @property
    def category_name(self):
        return self._name

    @property
    def openmmlabformat(self):
        return self._name

    @property
    def class_domain(self):
        return CLASSDOMAIN.get(self.domain_name)

    def __eq__(self, other):
        if self.domain_name != other.domain_name:
            return False
        if self.name != other.name:
            return False
        if len(self.parents) != len(self.parents):
            return False
        for p1 in self.parents:
            for p2 in other.parents:
                if p1 != p2:
                    return False
        return True

    def visualize(self, image, palette, **kwargs):
        draw_obj = ImageDraw.Draw(image)
        if self.category_name not in palette:
            palette[self.category_name] = tuple(np.random.randint(0, 255, size=[3]))
        color = palette[self.category_name]
        font = ImageFont.load_default()
        label_size = draw_obj.textsize(self.category_name, font)
        if "bbox" in kwargs:
            coords = np.array([[item.xyxy[0], item.xyxy[1] + 0.2 * label_size[1]] for item in kwargs["bbox"].values()])
        elif "polygon" in kwargs:
            coords = np.array([[item.point_for_draw[0], item.point_for_draw[1] + 0.2 * label_size[1]] for item in
                               kwargs["polygon"].values()])
        else:
            coords = np.array([[0, 0.2 * label_size[1]]])
        for coord in coords:
            draw_obj.rectangle([tuple(coord), tuple(coord + label_size)], fill=(*color, 255))
            draw_obj.text(tuple(coord), self.category_name, fill=(255, 255, 255, 255), font=font)
        del draw_obj
        return image

    def __repr__(self):
        return self.category_name


class LabelList(BaseGeometry):

    def __init__(self, label_list):
        self._label_list = list(label_list)

    @property
    def names(self):
        return [label.category_name for label in self._label_list]

    @property
    def category_names(self):
        return [label.category_name for label in self._label_list]

    @property
    def label_list(self):
        return self._label_list

    @property
    def class_domains(self):
        return [label.class_domain for label in self._label_list]

    def visualize(self, image, palette, **kwargs):
        draw_obj = ImageDraw.Draw(image)
        font = ImageFont.load_default()
        y_offset = np.zeros((1, 2))
        for label_obj in self.label_list:

            category_name = label_obj.category_name

            if category_name not in palette:
                palette[category_name] = tuple(np.random.randint(0, 255, size=[3]))
            color = palette[category_name]

            label_size = draw_obj.textsize(category_name, font)
            if "bbox" in kwargs:
                # coords.shape = [num_box, 2]
                coords = y_offset + np.array(
                    [[item.xyxy[0], item.xyxy[1] + 0.2 * label_size[1]] for item in kwargs["bbox"].values()])
            elif "polygon" in kwargs:
                coords = y_offset + np.array(
                    [[item.point_for_draw[0], item.point_for_draw[1] + 0.2 * label_size[1]] for item in
                     kwargs["polygon"].values()])
            elif "image_label_list" not in kwargs:
                coords = y_offset + np.array([[0, 0.2 * label_size[1]]])
            else:
                coords = []
                kwargs["image_label_list"].append(label_obj)

            y_offset += np.array([[0., 1.2 * label_size[1]]])
            for coord in coords:
                draw_obj.rectangle([tuple(coord), tuple(coord + label_size)], fill=(*color, 255))
                draw_obj.text(tuple(coord), category_name, fill=(255, 255, 255, 255), font=font)

        del draw_obj
        return image



from .field import Field
from ..geometry import Attributes, STRUCT
from ..exception import ValidationError
from ..warning import FieldNotFoundWarning


class StructMetaclass(type):
    def __new__(mcs, name, bases, attributes):
        super_new = super().__new__

        # Also ensure initialization is only performed for subclasses of Model
        # (excluding Model class itself).
        parents = [b for b in bases if isinstance(b, StructMetaclass)]
        if not parents:
            return super_new(mcs, name, bases, attributes)

        required = dict()  # 所有必须填写的field对象
        optional = dict()  # 所有可不填写的field对象
        attr = dict()  # 所有是attribute的field对象
        mappings = dict()  # 所有的field对象

        for k, v in attributes.items():
            if isinstance(v, Field):
                mappings[k] = v
                if v.is_attr:
                    attr[k] = v
                if v.is_optional:
                    optional[k] = v
                else:
                    required[k] = v

        for k in required.keys():
            attributes.pop(k)
        for k in optional.keys():
            attributes.pop(k)

        attributes["__required__"] = required
        attributes["__attr__"] = attr
        attributes["__optional__"] = optional
        attributes["__mappings__"] = mappings

        new_class = super_new(mcs, name, bases, attributes)
        STRUCT.register(name, new_class)
        return new_class


class Struct(dict, metaclass=StructMetaclass):
    def __init__(self, file_reader=None, **kwargs):

        super().__init__()
        self.attributes = Attributes()
        self.file_reader = file_reader
        self._keys = []
        if file_reader is not None:  # 说明是在赋值
            for k in self.__required__:
                if k not in kwargs:
                    FieldNotFoundWarning(f"Required field {k} is missing.")
                    continue
                setattr(self, k, kwargs[k])
                if k not in self.__attr__:
                    self._keys.append(k)
        for k in self.__optional__:
            if k in kwargs:
                setattr(self, k, kwargs[k])
                if k not in self.__attr__:
                    self._keys.append(k)

        self["$attributes"] = self.attributes
        self._keys.append("$attributes")

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(r"'Model' object has no attribute '%s'" % key)

    def __setattr__(self, key, value):
        if key not in self.__mappings__:
            self[key] = value
            return

        if hasattr(self.__mappings__[key], "set_file_reader"):
            self.__mappings__[key].set_file_reader(self.file_reader)
        try:
            if key in self.__attr__:
                self.attributes[key] = self.__mappings__[key].validate(value)
                return
            self[key] = self.__mappings__[key].validate(value)
        except ValidationError as error:
            raise ValidationError(f"Field '{key}' validation error: {error}.")

    def get_mapping(self):
        return self.__mappings__

    def keys(self):
        return tuple(self._keys)



# Generated by the dsdl parser. DO NOT EDIT!

from dsdl.types import *
from enum import Enum, unique


class Cifar10ImageClassificationClassDom(ClassDomain):
    Classes = [
        Label("airplane"),
        Label("automobile"),
        Label("bird"),
        Label("cat"),
        Label("deer"),
        Label("dog"),
        Label("frog"),
        Label("horse"),
        Label("ship"),
        Label("truck"),
    ]


class TemplateClassification(Struct):
    image = ImageField()
    label = LabelField(dom=Cifar10ImageClassificationClassDom, optional=True)
    confidence = NumField(optional=True)
    is_crowd = BoolField(optional=True, is_attr=True)



from .field import Field
from ..geometry import ImageMedia, SegmentationMap


class FileReader(object):
    def __init__(self, file_reader, args):
        self._file_reader = file_reader
        self.args = args

    def read(self):
        reader = self._file_reader
        with reader.load(self.args["$loc"]) as f:
            return f.read()


class UnstructuredObjectField(Field):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._file_reader = None

    def set_file_reader(self, file_reader):
        self._file_reader = file_reader

    @property
    def file_reader(self):
        return self._file_reader


class ImageField(UnstructuredObjectField):
    def __init__(self, *args, **kwargs):
        super(ImageField, self).__init__(*args, **kwargs)

    def validate(self, value):
        if isinstance(value, str):
            value = {"$loc": value}
        return ImageMedia(value["$loc"], FileReader(self.file_reader, value))


class SegMapField(UnstructuredObjectField):
    def __init__(self, dom, *args, **kwargs):
        super(SegMapField, self).__init__(*args, **kwargs)
        self.dom = dom

    def validate(self, value):
        if isinstance(value, str):
            value = {"$loc": value}
        return SegmentationMap(value['$loc'], FileReader(self.file_reader, value), self.dom)



import re
import networkx as nx
from typing import Dict, List


def sanitize_variable_name(varStr: str) -> str:
    """
    1. 将`.`替换为`__` 2.将（非字母开头）和（非字母数字及下划线）替换为`_`
    eg. apple.fruit_and_vegetables会转化为apple__fruit_and_vegetables
    """
    temp = varStr.split(".")
    temp = [re.sub("\W|^(?=\d)", "_", i) for i in temp]
    temp = "__".join(temp)
    return temp


def rreplace(s, old, new, occurrence):
    """
    从右向左的替换函数，类似replace,不过是反着的
    """
    li = s.rsplit(old, occurrence)
    return new.join(li)


def add_key_value_2_struct_field(field: str, key: str, value) -> str:
    """
    add key, value to field in struct,
    eg: key: optional, value: True, field: NumField(is_attr=True)
    return: NumField(optional=True, is_attr=True)
    """
    p = re.compile(r"[(](.*)[)]", re.S)  # 贪婪匹配
    k_v_list = re.findall(p, field)[0].strip()
    if k_v_list:
        k_v_list = k_v_list.split(",")
    else:
        k_v_list = []
    k_v_list.append(str(key) + "=" + str(value))
    temp = "(" + ", ".join(k_v_list) + ")"
    return field.replace("(" + re.findall(p, field)[0] + ")", temp)


def sort_nx(
    dict_sort_key: Dict[str, List[str]],
) -> List:
    """
    利用有向图对嵌套结构进行排序
    Args:
        dict_sort_key： {当前节点：[父节点],...}
    Returns: 排好序的节点的list，从父到子
    """
    define_graph = nx.DiGraph()
    define_graph.add_nodes_from(dict_sort_key.keys())
    for key, val in dict_sort_key.items():
        for base in val:
            for k in dict_sort_key.keys():
                if k in base:
                    define_graph.add_edge(k, key)
    if not nx.is_directed_acyclic_graph(define_graph):
        raise "define cycle found."
    ordered_keys = list(nx.topological_sort(define_graph))
    return ordered_keys



import torch.utils.data
from ..types import Struct
from ..geometry import STRUCT
from dsdl.dataset.utils import Util
import dsdl.objectio as objectio
from typing import List, Dict, Any, Callable, Optional, Union

try:
    from yaml import CSafeLoader as YAMLSafeLoader
except ImportError:
    from yaml import SafeLoader as YAMLSafeLoader


class Dataset(torch.utils.data.Dataset):
    # PALETTE用来存储每个类别的颜色（用于可视化）
    PALETTE = {}

    def __init__(
            self,
            samples: List[Dict[str, Any]],
            sample_type: Union[str, Struct],
            location_config: dict,
            pipeline: Optional[Callable[[Dict], Dict]] = None,
    ):
        self.location_config = location_config
        self.pipeline = pipeline  # 处理样本的函数
        self.samples = samples  # 样本所在的yaml文件路径
        sample_type = Util.extract_sample_type(sample_type)
        if isinstance(sample_type, str):
            self.sample_type = STRUCT.get(sample_type)
        else:
            self.sample_type = sample_type
        self.file_reader = self._load_file_reader(location_config)  # 样本的路径配置（如本地路径或是阿里云路径）
        self.sample_list = self._load_sample()  # 将yaml文件中的样本内容加载到self.sample_list中

    @staticmethod
    def _load_file_reader(config):
        config = config.copy()
        type_ = config.pop("type")
        cls = getattr(objectio, type_)
        try:
            reader = cls(**config)
        except Exception as e:
            print(f"raise exception {e} whe parse the location config {config}.")
            raise e
        return reader

    def _load_sample(self):
        """
        该函数的作用是将yaml文件中的样本转换为Struct对象，并存储到sample_list列表中
        """
        sample_list = []
        for sample in self.samples:
            sample_list.append(self._parse_struct(self.sample_type(file_reader=self.file_reader, **sample)))
        return sample_list

    def _parse_struct(self, sample):
        if isinstance(sample, Struct):
            data_item = {}
            struct_mapping = sample.get_mapping()
            for key in sample.keys():
                if key.startswith("$"):  # attributes
                    field_key = key
                    key = key.replace("$", "")
                    if field_key in data_item:
                        data_item[field_key][key] = getattr(sample, field_key)
                    else:
                        data_item[field_key] = {key: getattr(sample, field_key)}
                else:
                    field_key = Util.extract_key(struct_mapping[key])
                    if field_key in data_item:
                        data_item[field_key][key] = self._parse_struct(getattr(sample, key))
                    else:
                        data_item[field_key] = {key: self._parse_struct(getattr(sample, key))}
            return data_item

        elif isinstance(sample, list):
            return [self._parse_struct(item) for item in sample]
        else:
            return sample

    def process_sample(self, sample):
        return sample

    def __len__(self):
        return len(self.sample_list)

    def __getitem__(self, idx):
        sample = self.sample_list[idx]
        data = self.process_sample(sample)
        if self.pipeline is not None:
            data = self.pipeline(data)
        return data

    def get_sample_list(self):
        return self.sample_list

    @property
    def config(self):
        return self.location_config

    @staticmethod
    def format_sample(sample):
        raise NotImplementedError



from typing import Set, List, Dict, Optional
from dataclasses import dataclass, field
import re
from dsdl.exception import DefineSyntaxError
from collections import defaultdict
from .utils import sort_nx


@dataclass()
class SingleStructParam:
    struct_name: str
    params_dict: Dict[str, str] = field(default_factory=dict)
    parents_struct: List[str] = field(default_factory=list)


@dataclass()
class StructParams:
    params_list: List[SingleStructParam] = field(default_factory=list)


class ParserParam:
    def __init__(self, data_type, struct_defi):
        self.sample_param_map = self._parse_param_sample_type(raw=data_type)
        self.general_param_map = self._get_params(class_defi=struct_defi)

    @staticmethod
    def _parse_param_sample_type(raw: str) -> Optional[SingleStructParam]:
        """
        raw: ObjectDetectionSample[cdom=COCOClassFullDom]
        """
        c_dom = re.findall(r"\[(.*)\]", raw)
        if c_dom:
            data_sample_type = raw.replace("[" + c_dom[0] + "]", "", 1).strip()
            sample_param_map = SingleStructParam(struct_name=data_sample_type)
            temp = c_dom[0].split(",")
            for param in temp:
                param = param.strip()
                temp = param.split("=")
                sample_param_map.params_dict.update({temp[0].strip(): temp[1].strip()})
            return sample_param_map
        else:
            return None

    def _get_params(self, class_defi: Dict[str, Dict]) -> Dict[str, SingleStructParam]:
        self.general_param_map = defaultdict(SingleStructParam)
        ###################################################################################
        # 1。对class_defi循环，先拿到每个struct的params(self.general_param_map), 需要一个单独的循环，因为定义的顺序不一定
        # 以白皮书实例演练：计算机视觉：带场景分类的目标检测为列：
        # 先得到self.general_param_map =
        # {SceneAndObjectSample:{struct_name: SceneAndObjectSample
        #     params_dict: {scenedom: None, objectdom: None}
        #     parents_struct: []},
        # LocalObjectEntry:{struct_name: SceneAndObjectSample
        #     cdom: {cdom: None}
        #     parents_struct: []},}
        for define_name, define_value in class_defi.items():
            if "$def" in define_value and define_value["$def"] == "struct":
                struct_params = define_value.get("$params", None)
                if struct_params:
                    temp = SingleStructParam(
                        struct_name=define_name,
                        params_dict={key: None for key in struct_params},
                    )
                    self.general_param_map[define_name] = temp
        ###################################################################################
        # 2。然后填充`SingleStructParam`中的params_dict和parents_struct元素，
        # self.general_param_map =
        # {SceneAndObjectSample:{struct_name: SceneAndObjectSample
        #     params_dict: {scenedom: SceneDom, objectdom: ObjectDom}
        #     parents_struct: []},
        # LocalObjectEntry:{struct_name: SceneAndObjectSample
        #     cdom: {cdom: $scenedom}
        #     parents_struct: [SceneAndObjectSample]},}
        if len(self.general_param_map) == 0:
            return None
        if len(self.general_param_map) == 1:
            temp_param_map = list(self.general_param_map.values())[0]
            temp_name = list(self.general_param_map.keys())[0]
            for key in temp_param_map.params_dict.keys():
                try:
                    self.general_param_map[temp_name].params_dict[
                        key
                    ] = self.sample_param_map.params_dict[key]
                except KeyError as e:
                    raise DefineSyntaxError(f"miss the params {e} in definition")
        else:
            for define_name, define_value in class_defi.items():
                if "$def" in define_value and define_value["$def"] == "struct":
                    struct_params = define_value.get("$params", None)
                    if struct_params:
                        if define_name == self.sample_param_map.struct_name:
                            for key in self.general_param_map[define_name].params_dict:
                                try:
                                    self.general_param_map[define_name].params_dict[
                                        key
                                    ] = self.sample_param_map.params_dict[key]
                                except KeyError as e:
                                    raise DefineSyntaxError(
                                        f"miss the params {e} in definition"
                                    )

                        for raw_type in define_value["$fields"].values():
                            field_type = raw_type.replace(" ", "")
                            # 得到中括号中的 [...] 中的东西，贪婪匹配
                            for structure in self.general_param_map.keys():
                                if structure in field_type:
                                    self.general_param_map[
                                        structure
                                    ].parents_struct.append(define_name)
                                    try:
                                        # 如List[LocalObjectEntry[cdom=$objectdom]]会提取到'cdom=$objectdom'
                                        temp = re.findall(
                                            r"%s\[(.*?)\]" % str(structure), field_type
                                        )[0]
                                    except KeyError:
                                        raise DefineSyntaxError(
                                            f"definition error of filed {field_type}"
                                        )
                                    for param in temp.split(","):
                                        param = param.split("=")
                                        if len(param) != 2:
                                            raise DefineSyntaxError(
                                                f"error in params definition {field_type}"
                                            )
                                        key, value = param[0], param[1]
                                        try:
                                            self.general_param_map[
                                                structure
                                            ].params_dict[key] = value
                                        except KeyError as e:
                                            raise DefineSyntaxError(
                                                f"miss the params {e} in definition"
                                            )
                                    break
            ###################################################################################
            # 3。最后利用有向图对struct_name进行排序（父->子），本例中是[SceneAndObjectSample, LocalObjectEntry]
            # sort_param_dict = {SceneAndObjectSample: [], LocalObjectEntry:[SceneAndObjectSample]}
            sort_param_dict = {}
            for key, val in self.general_param_map.items():
                if len(val.parents_struct) > 1:  # 目前只能处理只有一个父节点的情况
                    raise DefineSyntaxError(f"error in definition")
                else:
                    sort_param_dict[key] = val.parents_struct  # {list[str]}
            ordered_keys = sort_nx(sort_param_dict)  # 用有向图排序
            # 对排序后的struct，按顺序填充参数的具体值，最后得到
            # self.general_param_map =
            # {SceneAndObjectSample:{struct_name: SceneAndObjectSample
            #     params_dict: {scenedom: SceneDom, objectdom: ObjectDom}
            #     parents_struct: []},
            # LocalObjectEntry:{struct_name: SceneAndObjectSample
            #     cdom: {cdom: ObjectDom}
            #     parents_struct: [SceneAndObjectSample]},}
            for struct in ordered_keys[1:]:
                parent_struct = self.general_param_map[struct].parents_struct
                if not parent_struct:
                    raise DefineSyntaxError(
                        f"each struct must have one parent struct, but {struct} can have no"
                    )
                if len(parent_struct) > 1:
                    raise DefineSyntaxError(
                        "each struct must have one parent struct, but can have more than one child struct.\n"
                        f"{struct} have more than one parent struct"
                    )
                parent_struct = parent_struct[0]
                for key, val in self.general_param_map[struct].params_dict.items():
                    if val.startswith("$"):
                        parent_key = val.replace("$", "").strip()
                        self.general_param_map[struct].params_dict[
                            key
                        ] = self.general_param_map[parent_struct].params_dict[
                            parent_key
                        ]
        return self.general_param_map

    def validate_params(self, struct_params_field: Set, struct_name: str) -> Set:
        input_params = set(self.general_param_map[struct_name].params_dict.keys())
        if input_params != struct_params_field:
            raise DefineSyntaxError(
                f"error of definition of params {struct_params_field}"
            )
        return struct_params_field



from .label import Label
from .registry import CLASSDOMAIN, LABEL
from .class_domain_attributes import Skeleton

CLASSDOMAIN_ATTRIBUTES = {
    "Skeleton": Skeleton,
}


class ClassDomainMeta(type):
    def __new__(mcs, name, bases, attributes):
        super_new = super().__new__
        parents = [b for b in bases if isinstance(b, ClassDomainMeta)]
        if not parents:
            return super_new(mcs, name, bases, attributes)

        classes = attributes.pop('Classes', [])
        classes = [_ for _ in classes if isinstance(_, Label)]
        mapping = {}
        for label in classes:
            label.set_domain(name)
            mapping[label.name] = label
            LABEL.registry(label)

        attributes["__mapping__"] = mapping
        attributes["__list__"] = classes
        attr_dic = {attr_k: CLASSDOMAIN_ATTRIBUTES[attr_k](attributes.pop(attr_k)) for attr_k in
                                        CLASSDOMAIN_ATTRIBUTES if attr_k in attributes}
        for attr_k in attr_dic:
            attr_dic[attr_k].set_domain(name)
        attributes["__attributes__"] = attr_dic


        new_cls = super_new(mcs, name, bases, attributes)
        CLASSDOMAIN.register(name, new_cls)
        return new_cls

    def __contains__(cls, item):
        container = getattr(cls, "__mapping__")
        if isinstance(item, str):
            return item in container
        elif hasattr(item, "category_name"):
            return item.category_name in container
        else:
            return False

    def __len__(cls):
        container = getattr(cls, "__list__")
        return len(container)


class ClassDomain(metaclass=ClassDomainMeta):
    @classmethod
    def get_labels(cls):
        return getattr(cls, "__list__")

    @classmethod
    def get_label_names(cls):
        labels = cls.get_labels()
        return [_.name for _ in labels]

    @classmethod
    def get_label(cls, name):
        if isinstance(name, str):
            container = getattr(cls, "__mapping__")
            if name in container:
                return container[name]
            else:
                raise KeyError(f"`{cls.__name__}` Domain doesn't have `{name}` category.")
        elif isinstance(name, int):
            container = getattr(cls, "__list__")
            if 1 <= name <= len(container):
                return container[name - 1]
            else:
                raise IndexError(f"There are only {len(container)} categories in `{cls.__name__}` domain.")
        else:
            raise RuntimeError(f"Invalid key {name}, only int/str keys are permitted.")

    @classmethod
    def get_attribute(cls, attr_name):
        attr_dic = getattr(cls, "__attributes__")
        return attr_dic.get(attr_name, None)



#!/usr/bin/env python3
#
# Copyright DSDL Team of OpenDatalab. All rights reserved.
#

"""Setup file."""

from setuptools import setup, find_packages


def readme():
    with open('README.md', encoding='utf-8') as f:
        content = f.read()
    return content


version = {}
with open("dsdl/__version__.py") as version_file:
    exec(version_file.read(), version)

setup(
    name="dsdl",
    version=version["__version__"],
    description="Python SDK for DSDL",
    long_description=readme(),
    long_description_content_type='text/markdown',
    author="DSDL team",
    author_email="dsdl-team@pjlab.org.cn",
    packages=find_packages(),
    package_data={},
    include_package_data=True,
    install_requires=[
        "pyyaml>=6.0",
        "click>=8.1.3",
        "networkx>=2.8.4",
        "Pillow>=9.1.0",
        "oss2>=2.15.0",
        "prettytable>=3.3.0",
        "opencv-python>=4.6.0.66",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires=">=3.8",
    entry_points={
        "console_scripts": [
            "dsdl = dsdl.cli:cli",
        ],
    },
)



from .registry import CLASSDOMAIN


class Skeleton:

    def __init__(self, skeleton, domain_name=None):
        self._value = skeleton
        self._domain_name = domain_name

    def set_domain(self, domain_name):
        if self._domain_name is None:
            self._domain_name = domain_name

    @property
    def domain_name(self):
        return self._domain_name

    @property
    def class_domain(self):
        return CLASSDOMAIN.get(self.domain_name)

    @property
    def value(self):
        return self._value

    def get_label_pairs(self):
        res = []
        for ind_pair in self._value:
            _p = []
            for ind in ind_pair:
                _p.append(self.class_domain.get_label(ind))
            res.append(_p)
        return res

    def get_point_pairs(self, keypoints):
        label_pairs = self.get_label_pairs()
        res = []
        for pair in label_pairs:
            _p = []
            for label in pair:
                _p.append(keypoints[label.name])
            res.append(_p)
        return res



class ValidationError(RuntimeError):
    pass


class ClassHasDefinedError(RuntimeError):
    pass


class ClassNotFoundError(RuntimeError):
    pass


class DefineSyntaxError(SyntaxError):
    pass


class DefineTypeError(TypeError):
    pass


class FileReadError(TypeError):
    pass





from .base_geometry import BaseGeometry

class Attributes(BaseGeometry):
    def __init__(self, **kwargs):
        self.container = {}
        for k, v in kwargs.items():
            self.container[k] = v

    def __setitem__(self, key, value):
        self.container[key] = value

    def __getitem__(self, item):
        return self.container[item]

    def __contains__(self, item):
        return item in self.container

    def keys(self):
        return self.container.keys()

    def values(self):
        return self.container.values()

    def __repr__(self):
        return self.container.__repr__()



import click
from abc import ABC, abstractmethod
from yaml import load as yaml_load
from dsdl.exception import DefineSyntaxError
from dsdl.warning import DuplicateDefineWarning
import os
from dataclasses import dataclass, field
from enum import Enum
from typing import Union, Optional
from collections import defaultdict
from .utils import *
from .parse_params import ParserParam
from .parse_field import ParserField, EleStruct
from .parse_class import ParserClass, EleClass

try:
    from yaml import CSafeLoader as YAMLSafeLoader
except ImportError:
    from yaml import SafeLoader as YAMLSafeLoader


class Parser(ABC):
    """
    定义一个Parser的抽象基类（没法实例化的）：
    用处： 强制子类实现某些方法：写框架时常用。
    """

    @abstractmethod
    def _parse(self, data_file: str, library_path: str):
        """
        将yaml文件中的模型（struct）和标签(label)部分校验之后读入某个变量中
        """
        pass

    @abstractmethod
    def _generate(self, output_file: str) -> Optional[str]:
        """
        将内存里面的模型（struct）和标签(label)部分输出成ORM模型（python代码）
        """
        pass

    def process(self, data_file, library_path, output_file):
        self._parse(data_file, library_path)
        dsdl_py = self._generate(output_file)
        print(
            f"Convert Yaml File to Python Code Successfully!\n"
            f"Yaml file (source): {data_file}\n"
            f"Output file (output): {output_file}"
        )
        return dsdl_py


class TypeEnum(Enum):
    CLASS_DOMAIN = "class_domain"
    STRUCT = "struct"


@dataclass()
class StructORClassDomain:
    name: str
    type: TypeEnum = TypeEnum.STRUCT
    field_list: List[Union[EleStruct, EleClass]] = field(default_factory=list)
    parent: List[str] = None


class DSDLParser(Parser, ABC):
    def __init__(self):
        self.struct_name = set()
        self.define_map = defaultdict(StructORClassDomain)
        self.dsdl_version = None
        self.meta = dict()

    def _parse(self, data_file: str, library_path: str):
        """
        将yaml文件中的模型（struct）和标签(label)部分校验之后读入变量self.define_map中
            input_file_list: 读入的yaml文件
        """
        with open(data_file, "r") as f:
            desc = yaml_load(f, Loader=YAMLSafeLoader)

        # 校验必须有meta和$dsdl-version字段，见白皮书2.1
        try:
            self.dsdl_version = desc["$dsdl-version"]  # 存版本号，后续应该会使用（目前木有用）
        except KeyError as e:
            raise DefineSyntaxError(f"data yaml must contains {e} section")
        try:
            self.meta = desc["meta"]  # 存版meta信息，后续应该会使用（目前木有用）
        except KeyError as e:
            raise DefineSyntaxError(f"data yaml must contains {e} section")

        # 校验必须有data字段和data中的sample-type字段
        try:
            data_sample_type = desc["data"]["sample-type"]
        except KeyError as e:
            raise DefineSyntaxError(f"data yaml must contains {e} in `data` section")

        if "$import" in desc:
            import_list = desc["$import"]
            import_list = [
                os.path.join(library_path, p.strip() + ".yaml") for p in import_list
            ]
        else:
            import_list = []
        if "defs" in desc:
            # 获取yaml中模型（struct）和标签(label)部分的内容，存储在变量class_defi中，
            # 因为有不同格式的yaml(数据和模型放同一个yaml中或者分开放)，所以用if...else分别做处理
            # 注意区分这个root_class_defi,为啥要把他先存好？如果import的yaml中，有重复的模型，需要用它覆盖，参见白皮书2.5.1
            root_class_defi = desc["defs"].items()
            self.dsdl_version = desc["$dsdl-version"]  # 存版本号，后续应该会使用（目前木有用）
        else:
            root_class_defi = dict()

        import_desc = dict()
        for input_file in import_list:
            with open(input_file, "r") as f:
                import_desc.update(yaml_load(f, Loader=YAMLSafeLoader))

        if "defs" in import_desc:
            # 获取yaml中模型（struct）和标签(label)部分的内容，存储在变量class_defi中，
            # 因为有不同格式的yaml(数据和模型放同一个yaml中或者分开放)，所以用if...else分别做处理
            class_defi = import_desc["defs"]
        else:
            class_defi = import_desc
        # root_class_defi是数据yaml里面定义的模型，如果和import里面的重复了，会覆盖掉前面import的。参见白皮书2.5.1
        class_defi.update(root_class_defi)

        # get self.data_sample_type and self.sample_param_map
        PARAMS = ParserParam(data_type=data_sample_type, struct_defi=class_defi)
        for define_name, define_value in class_defi.items():
            if define_name.startswith("$"):
                continue
            try:
                define_type = define_value["$def"]
            except KeyError as e:
                raise DefineSyntaxError(
                    f"{define_name} section must contains {e} sub-section"
                )
            if define_type == "struct" or define_type == "class_domain":
                if define_name in self.struct_name:
                    raise DuplicateDefineWarning(f"{define_name} has defined.")
                self.struct_name.add(define_name)

        # loop for `class_defi` section，deal with each `struct` and `class_domain`
        for define_name, define_value in class_defi.items():
            if define_name.startswith("$"):
                # skip section like: $dsdl-version
                continue

            # each yaml file must contain '$def' section
            define_type = define_value["$def"]

            if define_type == "struct":
                define_info = StructORClassDomain(name=define_name)
                define_info.type = TypeEnum.STRUCT
                FIELD_PARSER = ParserField(self.struct_name)
                # verify each ele of `struct` in `class_defi`, and save in define_info
                struct_params = define_value.get("$params", None)
                field_list = dict()
                for raw_field in define_value["$fields"].items():
                    field_name = raw_field[0].strip()
                    field_type = raw_field[1].strip()
                    if not field_name.isidentifier():
                        err = (
                            f"'{field_name}' must be a a valid identifier. "
                            f"Field name is considered a valid identifier if "
                            f"it only contains alphanumeric letters (a-z) and (0-9), or underscores (_). "
                            f"A valid identifier cannot start with a number, or contain any spaces."
                        )
                        raise DefineSyntaxError(err)
                    if struct_params:
                        for param, value in PARAMS.general_param_map[
                            define_name
                        ].params_dict.items():
                            field_type = field_type.replace("$" + param, value)
                    field_list[field_name] = EleStruct(
                        name=field_name,
                        type=FIELD_PARSER.pre_parse_struct_field(
                            field_name, field_type
                        ),
                    )
                # deal with `$optional` section after `$fields` section，
                # because we must ensure filed in `$optional` is the `filed_name` in `$fields` section.
                if "$optional" in define_value or FIELD_PARSER.optional:
                    optional_set = (
                        set(define_value["$optional"]) | FIELD_PARSER.optional
                    )
                    for optional_name in optional_set:
                        optional_name = optional_name.strip()
                        if optional_name in field_list:
                            temp_type = field_list[optional_name].type
                            temp_type = add_key_value_2_struct_field(
                                temp_type, "optional", True
                            )
                            field_list[optional_name].type = temp_type
                        else:
                            raise DefineSyntaxError(f"{optional_name} is not in $field")
                for attr_name in FIELD_PARSER.is_attr:
                    temp_type = field_list[attr_name].type
                    temp_type = add_key_value_2_struct_field(temp_type, "is_attr", True)
                    field_list[attr_name].type = temp_type

                # get processed struct filed and save it in define_info
                if not field_list:
                    raise DefineSyntaxError(
                        "Struct must have fields more than or equal to 1"
                    )
                define_info.field_list = list(field_list.values())

            elif define_type == "class_domain":
                CLASS_PARSER = ParserClass(define_name, define_value["classes"])
                define_info = StructORClassDomain(name=CLASS_PARSER.class_name)
                # verify each ele (in other words: each label) of `class_domain`, and save in define_info
                define_info.type = TypeEnum.CLASS_DOMAIN
                define_info.field_list = CLASS_PARSER.class_field
                define_info.parent = CLASS_PARSER.super_class_list
            else:
                raise DefineSyntaxError(
                    f"error type {define_type} in yaml, type must be class_dom or struct."
                )

            self.define_map[define_info.name] = define_info

    def _generate(self, output_file: str = None) -> Optional[str]:
        """
        将内存里面的模型（struct）和标签(label)部分输出成ORM模型（python代码）
        """
        # check define cycles. 如果有环形（就是循环定义）那是不行滴～
        define_graph = nx.DiGraph()
        define_graph.add_nodes_from(self.define_map.keys())
        for key, val in self.define_map.items():
            if val.type == TypeEnum.STRUCT:
                for field_list in val.field_list:
                    for k in self.define_map.keys():
                        if k in field_list.type:
                            define_graph.add_edge(k, key)
            elif val.type == TypeEnum.CLASS_DOMAIN:
                for field_list in val.parent:
                    for k in self.define_map.keys():
                        if k in field_list:
                            define_graph.add_edge(k, key)
        if not nx.is_directed_acyclic_graph(define_graph):
            raise "define cycle found."

        dsdl_py = "# Generated by the dsdl parser. DO NOT EDIT!\n"
        dsdl_py += "from dsdl.types import *\n\n\n"
        ordered_keys = list(nx.topological_sort(define_graph))
        for idx, key in enumerate(ordered_keys):
            val = self.define_map[key]
            if val.type == TypeEnum.STRUCT:
                dsdl_py += f"class {key}(Struct):\n"
                for field_list in val.field_list:
                    dsdl_py += f"""    {field_list.name} = {field_list.type}\n"""
            if val.type == TypeEnum.CLASS_DOMAIN:
                dsdl_py += f"class {key}(ClassDomain):\n"
                dsdl_py += "    Classes = [\n"
                for ele_class in val.field_list:
                    if ele_class.super_categories:
                        temp = ", ".join(ele_class.super_categories)
                        dsdl_py += f"""        Label("{ele_class.label_value}", supercategories=[{temp}]),\n"""
                    else:
                        dsdl_py += f"""        Label("{ele_class.label_value}"),\n"""
                dsdl_py += "    ]\n"
            if idx != len(ordered_keys) - 1:
                dsdl_py += "\n\n"

        if output_file:
            with open(output_file, "w") as of:
                print(dsdl_py, file=of)
        else:
            return dsdl_py


def dsdl_parse(
    dsdl_yaml: str,
    dsdl_library_path: str = "dsdl/dsdl_library",
    output_file: str = None,
) -> Optional[str]:
    """
    Main function of parser yaml files to .py dsdl struct definition code.

    Arguments:
        dsdl_yaml: file path of `data definition yaml file`;
        dsdl_library_path: file path of '`$import` path' in `dsdl_yaml` file.
        output_file: output file path. if None, return string, else, generate .py file in output file path.

    Returns:
        Optional[str]: if output_file=None, return string of dsdl definition .py file;
                       else generate a .py file in `output_file` path.
    """
    dsdl_parser = DSDLParser()
    res = dsdl_parser.process(dsdl_yaml, dsdl_library_path, output_file)
    return res


@click.command()
@click.option(
    "-y",
    "--yaml",
    "dsdl_yaml",
    type=str,
    required=True,
)
@click.option(
    "-p",
    "--path",
    "dsdl_library_path",
    type=str,
    default="dsdl/dsdl_library",
)
def parse(dsdl_yaml: str, dsdl_library_path: str):
    """
    a separate cli tool function for user to parser yaml files to .py dsdl struct definition code.

    Arguments:
        dsdl_yaml: file path of `data definition yaml file`;
        dsdl_library_path: file path of '`$import` path' in `dsdl_yaml` file.

    Returns:
        None: generate a .py file in the same folder of `dsdl_yaml` file.
    """
    dsdl_name = os.path.splitext(os.path.basename(dsdl_yaml))[0]
    output_file = os.path.join(os.path.dirname(dsdl_yaml), f"{dsdl_name}.py")
    dsdl_parse(dsdl_yaml, dsdl_library_path, output_file)



from typing import Optional, Dict, Tuple, List
import os
from collections import defaultdict

from .parser import Parser
from ...geometry import LabelList, BBox
from copy import deepcopy


class VisualizerUtil:

    @staticmethod
    def extract_field_pipeline(result_dic, sample, field_name):
        info = Parser.flatten_sample(sample, f"${field_name}")
        result_dic[field_name] = info

    @staticmethod
    def whole_pipeline(field_lst, sample):
        result_dic = {}
        for field_name in field_lst:
            VisualizerUtil.extract_field_pipeline(result_dic, sample, field_name)
        return result_dic

    @staticmethod
    def sort_field(field_lst):
        field_lst = list(field_lst)
        # 定义画图时的先后顺序，数值越小的越先画
        field_order = {"bbox": 20, "polygon": 10, "label": 30, "keypoint": 15, "segmap": 0, "others": -1}

        field_lst = sorted(field_lst, key=lambda k: field_order.get(k, -1))
        return field_lst


class ImageVisualizer:
    def __init__(self, image, palette):
        self.image = image
        self.palette = palette
        self.ground_truths = defaultdict(dict)

    def append_ground_truth(self, gt_path, gt_item):
        """
        {
            ./object/1:{
                bbox: {bb1: bbox_obj, bb2: bbox_obj}
                label: [l1: label_obj, l2: label_obj]
                bool: []
            }

            ./object/2: {
                bbox: {}
            }
        }
        """
        gt_dir, gt_name = os.path.split(gt_path)
        field_key = gt_item.__class__.__name__.lower()
        if field_key not in self.ground_truths[gt_dir]:
            self.ground_truths[gt_dir][field_key] = {gt_name: gt_item}
        else:
            self.ground_truths[gt_dir][field_key][gt_name] = gt_item

    def visualize(self):
        image = self.image.to_image()
        image = image.convert("RGBA")
        image_label_lst = []
        for gt_dir, gt_item in self.ground_truths.items():
            field_keys = VisualizerUtil.sort_field(gt_item.keys())
            for field_key in field_keys:
                if field_key == "label":
                    image = LabelList(gt_item[field_key].values()).visualize(image=image,
                                                                             image_label_list=image_label_lst,
                                                                             palette=self.palette, **gt_item)
                else:
                    for ann_item in gt_item[field_key].values():
                        if hasattr(ann_item, "visualize"):
                            image = ann_item.visualize(image=image, palette=self.palette, **gt_item)
        LabelList(image_label_lst).visualize(image=image, palette=self.palette, bbox={"temp": BBox(0, 0, 0, 0)})
        return image

    def format(self):
        result = {"media": self.image, "ann": list(self.ground_truths.values())}
        return result


class ImageVisualizePipeline(VisualizerUtil):
    """
    step 1:

    该类用于处理Dataset中输出的样本，指定field_list后，会先调用方法 self.whole_pipeline，提取各个字段的信息，举例：
    Dataset中的某个sample形如：
        sample = {
        "$image": {"img2": img_obj, "img1":img_obj},
        "$list": {"objects": [{"$image": {"img1": img_obj}, "$bbox": {"box": box_obj}}],
                  "object2": [{"$bbox":{"box": box_obj}}, {"$bbox":{"box": box_obj}}], }
        }
    指定的field_list内容为：[bbox, image]
    则self.whole_pipeline方法会将sample处理为：
    {
        ”image": {
            "./img2": img_obj,
            "./img1": img_obj,
            "./objects/0/img1": img_obj,
        },
        "bbox": {
            "./objects/0/box": box_obj,
            "./object2/0/box": box_obj,
            "./object2/1/box": box_obj
        }
    }

    step2:
    然后会调用self.group_media_and_ann方法，将图像与与之”目录路径“相同的标注进行组合，组合后图像和对应的标注会存储在一个 ImageVisualizer
    对象当中。ImageVisualizer类中还定义了可视化的相关方法，从而可以对每个图像&标注组合进行可视化。方法的返回值是一个字典，字典的键为图像路径，
    值即为ImageVisualizer对象，形如：
    {
        "./img2": ImageVisualizer_obj,
        "./img1": ImageVisualizer_obj,
        "./objects/0/img1": ImageVisualizer_obj,
    }

    step3:
    调用format/visualize方法，从而可以对各个ImageVisualizer对象进行文本/图像可视化。
    """
    PALETTE = {}

    def __init__(self, field_list: List[str], sample, palette: Optional[Dict[str, Tuple]] = None):
        self.field_list = field_list
        self.palette = self.PALETTE
        if palette is not None:  # 如果用户自己定义了label调色盘，则更新默认的调色盘
            self.palette = palette
        if "image" not in field_list:  # 因为是imagevisualizer类，所以field_list中必须包含image字段
            raise ValueError("'image' field not found in the field list to be visualized.")
        self.data_dic = self.whole_pipeline(field_list, sample)
        self.visualize_result = self.group_media_and_ann()

    def group_media_and_ann(self):
        data_dic = deepcopy(self.data_dic)
        image_dic = data_dic.pop("image")
        image_paths = list(image_dic.keys())
        # 将图像按照路径进行从长到短排序
        image_paths = sorted(image_paths, key=lambda x: len(x.split("/")), reverse=True)
        result_dic = {}
        for image_path in image_paths:
            image_dir = os.path.split(image_path)[0]
            image_obj = image_dic[image_path]
            result_dic[image_path] = ImageVisualizer(image_obj, self.palette)
            for ann_dic in data_dic.values():
                pop_keys = []
                for ann_path, ann_obj in ann_dic.items():
                    if ann_path.startswith(image_dir):
                        result_dic[image_path].append_ground_truth(ann_path, ann_dic[ann_path])
                        pop_keys.append(ann_path)
                for pop_key in pop_keys:
                    ann_dic.pop(pop_key)
        return result_dic

    def visualize(self):
        vis_result = {}
        for image_path, image_vis_item in self.visualize_result.items():
            vis_result[image_path] = image_vis_item.visualize()
        return vis_result

    def format(self):
        format_result = []
        for image_vis_item in self.visualize_result.values():
            format_result.append(image_vis_item.format())
        return format_result
